//! Calibre library book discovery and metadata extraction.

use std::{
    fs::{self, File},
    io::{self, Read},
    num::ParseIntError,
    path::{Path, PathBuf},
};

use quick_xml::{Reader, encoding::EncodingError, events::Event};
use ring::digest::{Context, Digest, SHA256};
use snafu::{ResultExt, Snafu};
use walkdir::WalkDir;

/// A book found in a Calibre library.
#[derive(Debug)]
pub struct Book {
    /// Calibre book ID from metadata.
    pub id: i64,
    /// File extension (epub, pdf, etc).
    pub extension: String,
    /// Full path to the book file.
    pub path: PathBuf,
    /// SHA256 hash of the book file.
    pub hash: Vec<u8>,
}

/// Iterator over books in a Calibre library directory structure.
pub struct BookIterator {
    /// Directory walker for traversing the library.
    walker: walkdir::IntoIter,
}

/// Error when computing file hash.
#[derive(Debug, Snafu)]
#[snafu(display("Failed to delete old document versions."))]
pub struct HashError {
    source: io::Error,
}

/// Errors that occur when reading Calibre ID from OPF metadata.
#[derive(Debug, Snafu)]
pub enum ReadCalibreIdError {
    #[snafu(display("Failed to read opf xml."))]
    ReadXml { source: quick_xml::Error },
    #[snafu(display("Failed to read xml text event."))]
    ReadXmlText { source: quick_xml::Error },
    #[snafu(display("Failed to decode xml text."))]
    DecodeXml { source: EncodingError },
    #[snafu(display("Failed to parse calibre id as integer."))]
    ParseId { source: ParseIntError },
    #[snafu(display("Failed to find calibre id in opf."))]
    CalibreIdMissing,
}

/// Error when iterating through library directories.
#[derive(Debug, Snafu)]
#[snafu(display("Failed to walk the library tree."))]
pub struct IterateError {
    source: walkdir::Error,
}

impl BookIterator {
    /// Create a new iterator for the given Calibre library root directory.
    pub fn new(root_path: &PathBuf) -> Self {
        Self {
            walker: WalkDir::new(root_path)
                .follow_links(true)
                .max_depth(3) // root/author/book/files
                .into_iter(),
        }
    }

    /// Check if a file path represents a supported ebook format.
    fn is_ebook(path: &Path) -> bool {
        let extensions = ["epub", "pdf"];
        path.extension()
            .and_then(|ext| ext.to_str())
            .map(|ext| extensions.contains(&ext.to_lowercase().as_str()))
            .unwrap_or(false)
    }
}

/// Compute SHA256 hash of data from a reader.
fn sha256_digest<R: Read>(mut reader: R) -> Result<Digest, HashError> {
    let mut context = Context::new(&SHA256);
    let mut buffer = [0; 1024];

    loop {
        let count = reader.read(&mut buffer).context(HashSnafu)?;
        if count == 0 {
            break;
        }
        context.update(&buffer[..count]);
    }

    Ok(context.finish())
}

/// Extract the Calibre book ID from OPF metadata XML.
fn read_calibre_id(opf: &str) -> Result<i64, ReadCalibreIdError> {
    let mut reader = Reader::from_str(opf);
    reader.config_mut().trim_text(true);

    loop {
        match reader.read_event() {
            Err(e) => return Err(e).context(ReadXmlSnafu),
            Ok(Event::Eof) => break,
            Ok(Event::Start(e)) => {
                if e.name().as_ref() == b"dc:identifier"
                    && e.attributes().any(|attr| {
                        attr.ok()
                            .filter(|a| {
                                a.key.as_ref() == b"id" && a.value.as_ref() == b"calibre_id"
                            })
                            .is_some()
                    })
                {
                    if let Event::Text(text) = reader.read_event().context(ReadXmlTextSnafu)? {
                        let calibre_id = text.decode().context(DecodeXmlSnafu)?.into_owned();
                        let calibre_id = calibre_id.parse::<i64>().context(ParseIdSnafu)?;
                        return Ok(calibre_id);
                    }
                }
            }
            _ => (),
        }
    }

    CalibreIdMissingSnafu.fail()
}

impl Iterator for BookIterator {
    type Item = Result<Book, IterateError>;

    fn next(&mut self) -> Option<Self::Item> {
        loop {
            match self.walker.next()? {
                Ok(entry) => {
                    if entry.file_type().is_dir() {
                        let path = entry.path().to_owned();
                        let opf_path = path.join("metadata.opf");
                        if opf_path.exists() {
                            let files = fs::read_dir(&path).ok()?;
                            for f in files.flatten() {
                                let f = f.path();

                                if BookIterator::is_ebook(&f) {
                                    let opf = fs::read_to_string(opf_path).ok()?;
                                    let id = read_calibre_id(&opf).ok()?;

                                    let book = File::open(&f).ok()?;
                                    let hash = sha256_digest(book).ok()?.as_ref().to_vec();

                                    return Some(Ok(Book {
                                        id,
                                        extension: f.extension()?.to_string_lossy().to_string(),
                                        path: f,
                                        hash,
                                    }));
                                }
                            }
                        }
                    }
                }
                Err(e) => return Some(Err(e).context(IterateSnafu)),
            }
        }
    }
}
